import re
import gzip
import json
import traceback
import numpy as np
import SimpleITK as sitk

from pathlib import Path
from shutil import copyfile
from uuid import uuid4
from loguru import logger
# Stats tables that can be formatted for database insertion (excludes conversions, segmentations)
STATS_TABLES = [
    "metadata",
    "vol_int",
    "contrast",
    "vertebrae",
    "aorta_metrics",
    "aorta_diameters",
    "tissues_volumetric",
    "tissues_vertebral",
    "iliac",
]
from mircat_v2.dicom_conversion.converter import get_metadata


class Nifti:
    """A class used to represent a nifti file for mircat-v2 - including spacing, metadata, segmentations and others!"""

    def __init__(self, nifti_path: str | Path):
        nifti_path = Path(nifti_path).resolve()
        if not nifti_path.exists():
            message = f"Nifti file not found at {nifti_path}! Please check existence."
            logger.error(message)
            raise FileNotFoundError()
        if ".nii" not in nifti_path.suffixes:
            message = f"Input {nifti_path} not recognized as a nifti file. Please ensure file suffix is *.nii or *.nii.gz"
            logger.error(message)
            raise NotNiftiFileError()
        self.path = nifti_path
        self.parent = nifti_path.parent
        self.name = nifti_path.with_suffix("").stem
        self.seg_folder = self.parent / f"{self.name}_segs"
        self.segmentations = list(self.seg_folder.glob(f"{self.name}*.nii.gz"))
        try:
            self.img = sitk.ReadImage(nifti_path)
        except Exception as e:
            logger.error(
                f"Error loading {nifti_path} with SimpleITK :{e}\n{traceback.format_exc()}"
            )
            raise SimpleITKReadError()
        self.spacing = self.img.GetSpacing()
        self.shape = self.img.GetSize()
        try:
            if (self.parent / "metadata.json").exists():
                logger.trace(
                    f"{self.parent / 'metadata.json'} found. Loading metadata."
                )
                with (self.parent / "metadata.json").open() as f:
                    self.metadata: dict = json.load(f)
                if self.metadata.get("series_uid") is None:
                    self.metadata["series_uid"] = str(uuid4())
                    self._write_metadata()

            elif (self.parent / "header_info.json").exists():
                logger.trace(
                    f"Old format metadata {self.parent / 'header_info.json'} found. Loading metadata."
                )
                with (self.parent / "header_info.json").open() as f:
                    self.metadata: dict = json.load(f)
                logger.trace("Updating metadata in Nifti.__init__")
                self.update_metadata_file()
            else:
                logger.debug(
                    f"No metadata.json or header_info.json file found next to {nifti_path}. Setting metadata['series_uid'] to a file with a generated uuid."
                )
                self.metadata: dict = {"series_uid": str(uuid4())}
                self._write_metadata()
        except json.JSONDecodeError as e:
            logger.warning(
                "Existing metadata cannot be loaded from json. Creating a generated Series UID"
            )
            self.metadata: dict = {"series_uid": str(uuid4())}
            self._write_metadata()

    def update_metadata_file(self) -> None:
        dicoms = list(self.parent.glob("*.dcm"))
        if not dicoms:
            logger.trace(
                f"No .dcm extension file found in {self.parent} - looking for dicom without extension"
            )
            dicoms = [
                f
                for f in self.parent.iterdir()
                if f.is_file() and re.match(r"^[\d.]+$", f.name)
            ]
            if not dicoms:
                logger.trace(f"No reference dicom found in {self.parent}")
                if self.metadata.get("series_number") is None:
                    self.metadata["series_number"] = self.name.split("_")[0]
                self.metadata["series_uid"] = str(uuid4())
                # Delete the old version
        if dicoms:
            self.metadata = get_metadata(dicoms[0])
            if self.metadata.get("series_uid") is None:
                self.metadata["series_uid"] = str(uuid4())
        # (self.parent / 'header_info.json').unlink()
        self._write_metadata()

    def _write_metadata(self):
        with (self.parent / "metadata.json").open("w") as f:
            json.dump(self.metadata, f, indent=2)


class S3SegNifti:
    def __init__(self, nifti_path: str | Path, temp_dir: Path):
        nifti_path = Path(nifti_path)
        self.name = nifti_path.with_suffix("").stem
        self.original_path = nifti_path
        self.original_seg_folder = nifti_path.parent / f"{self.name}_segs"
        self.parent = temp_dir / str(uuid4())
        self.parent.mkdir(exist_ok=True, parents=True)
        self.seg_folder = self.parent / f"{self.name}_segs"
        self.path = self.parent / self.original_path.name
        logger.debug(
            "copying {} to temporary location {}", self.original_path, self.path
        )
        try:
            copyfile(self.original_path, self.path)
        except Exception as e:
            logger.error(
                f"Error copying {self.path} to temporary location {self.temp_path} :{e}\n{traceback.format_exc()}"
            )
            raise FileNotFoundError()
        try:
            self.img = sitk.ReadImage(self.path)
        except Exception as e:
            logger.error(
                f"Error loading {self.path} with SimpleITK :{e}\n{traceback.format_exc()}"
            )
            raise SimpleITKReadError()
        self.spacing = self.img.GetSpacing()
        self.shape = self.img.GetSize()
        # We ignore metadata on s3 to save read time
        self.metadata = {}
        self.task_files = {}

    def resample_and_save_for_segmentation(
        self, new_spacing, output_path, interpolator_type
    ) -> None:
        if len(new_spacing) == 2:
            # Set the new spacing z-length to be the same as the original
            new_spacing = [*new_spacing, self.spacing[-1]]
        # Resample image - always not a label
        resampled = resample_with_sitk(
            self.img,
            new_spacing=new_spacing,
            is_label=False,
            interpolator_type=interpolator_type,
        )
        sitk.WriteImage(resampled, output_path)
        logger.debug(
            "Successfully resampled {} and wrote to {}", self.path, output_path
        )


class SegNifti(Nifti):
    def __init__(self, nifti_path: str | Path):
        super().__init__(nifti_path)
        self.task_files: dict = {}

    def resample_and_save_for_segmentation(
        self, new_spacing, output_path, interpolator_type
    ) -> None:
        if len(new_spacing) == 2:
            # Set the new spacing z-length to be the same as the original
            new_spacing = [*new_spacing, self.spacing[-1]]
        # Resample image - always not a label
        resampled = resample_with_sitk(
            self.img,
            new_spacing=new_spacing,
            is_label=False,
            interpolator_type=interpolator_type,
        )
        sitk.WriteImage(resampled, output_path)
        logger.debug(
            "Successfully resampled {} and wrote to {}", self.path, output_path
        )


class StatsNifti(Nifti):
    def __init__(self, nifti_path: str | Path, overwrite: bool = False):
        super().__init__(nifti_path)
        # TODO this needs to be updated to account for gzipping option
        stats_file = self.seg_folder / f"{self.name}_stats.json"
        self.stats_file = stats_file

        if stats_file.exists() and not overwrite:
            logger.trace(f"Stats file {stats_file} found. Loading statistics.")
            try:
                with stats_file.open() as f:
                    self.stats = json.load(f)
            except json.JSONDecodeError:
                logger.warning(
                    f"Error decoding JSON from {stats_file}. Initializing empty stats."
                )
                self.stats = {}
        else:
            if not overwrite:
                logger.trace(
                    f"No stats file found at {stats_file}. Setting stats to empty dictionary."
                )
            else:
                logger.trace(
                    f"Stats file {stats_file} exists but overwrite is set to True. Initializing empty stats."
                )
            self.stats = {}
        self.add_stats("nifti", str(self.path))
        # We want this initialized regardless of it's it's empty or not
        self.add_stats("metadata", self.metadata)
        if self.metadata != {}:
            logger.trace(f"Merging metadata with stats for {self.name}.")
            study_uid = self.metadata.get("study_uid")
            series_uid = self.metadata.get("series_uid")
            if series_uid is None:
                series_uid = self._generate_series_uid()
            self.add_stats("series_uid", series_uid)
            self.add_stats("study_uid", study_uid)
        # TODO - currently vertebrae midlines will load regardless of the desired stats resolution - which is not correct.
        # If the stats resolution is different from the nifti resolution, we must remeasure the midlines.
        if self.stats.get("vertebrae") is not None:
            self.set_vertebrae_midlines(self.stats["vertebrae"])
        # Placeholder for database stats
        self.db_stats: dict[str, list[dict]] = {}

    def set_task_to_id_map(self, task_to_id_map: dict) -> None:
        self.task_to_id_map = task_to_id_map
        return self

    def set_id_to_seg_map(self, id_to_seg_map: dict) -> None:
        """Set the mapping of task IDs to task names for this nifti."""
        self.id_to_seg_map = id_to_seg_map
        return self

    def set_vertebrae_midlines(self, vertebrae_stats: dict = None):
        """Set the vertebrae midlines for this nifti."""
        if vertebrae_stats is None:
            if "vertebrae" not in self.stats:
                logger.warning(
                    "No vertebrae stats found in nifti. Cannot set vertebrae midlines."
                )
                return self
            vertebrae_stats = self.stats["vertebrae"]
        vertebrae_midlines = {
            k: v.get("midline")
            for k, v in vertebrae_stats.items()
            if isinstance(v, dict)
        }
        self.vertebrae_midlines = vertebrae_midlines
        return self

    def add_stats(self, key: str, value: any, subdict=None) -> None:
        """Add a key-value pair to the stats dictionary."""
        if key in self.stats:
            logger.trace(
                f"Key {key} already exists in stats for {self.path}. Overwriting value."
            )
        if subdict:
            self.stats[subdict][key] = value
        else:
            self.stats[key] = value
        return self

    def preprocess_for_stats(
        self, new_spacing, image_resampler, label_resampler
    ) -> None:
        logger.trace(
            "Resampling image and segmentations for nifti {} to new spacing {} using {} for images and {} for labels",
            self.name,
            new_spacing,
            image_resampler,
            label_resampler,
        )
        self.img = resample_with_sitk(
            self.img,
            new_spacing=new_spacing,
            is_label=False,
            interpolator_type=image_resampler,
            reorient="LAI",
        )
        logger.trace("image resampled")
        self.spacing = self.img.GetSpacing()
        self.shape = self.img.GetSize()
        self.add_stats("stats_resolution", new_spacing)
        new_id_to_seg_map = {}
        for seg_id, seg_path in self.id_to_seg_map.items():
            label = sitk.ReadImage(seg_path)
            resampled_label = resample_with_sitk(
                label,
                is_label=True,
                interpolator_type=label_resampler,
                reference_image=self.img,
            )
            assert resampled_label.GetSize() == self.img.GetSize(), (
                f"Resampled label size {resampled_label.GetSize()} does not match image size {self.img.GetSize()}"
            )
            new_id_to_seg_map[seg_id] = resampled_label
        self.set_id_to_seg_map(new_id_to_seg_map)
        return self

    def check_for_segmentations(self, task_list, task_map):
        task_to_id_map = {}
        id_to_segmentation_map = {}
        # These are the mircat-v1 output file suffixes for these 3 tasks - we just have to remap to the new values
        old_to_new_map = {"total": "999", "tissues": "481", "body": "299", "998": "999"}
        for task in task_list:
            allowed_segmentations = task_map.get(task, [])
            for seg_id in allowed_segmentations:
                seg_file = f"{self.name}_{seg_id}.nii.gz"
                if seg_file in [seg.name for seg in self.segmentations]:
                    if old_to_new_map.get(seg_id, False):
                        logger.debug(f"Flipping {seg_id} to {old_to_new_map[seg_id]}")
                        seg_id = old_to_new_map[seg_id]
                    task_to_id_map[task] = seg_id
                    id_to_segmentation_map[seg_id] = self.seg_folder / seg_file
                    # We break because the task_map is sorted by preference
                    break
        logger.debug(f"{task_to_id_map=}")
        self.set_task_to_id_map(task_to_id_map)
        self.set_id_to_seg_map(id_to_segmentation_map)
        return self

    def save_json_stats(self, gzip_file: bool = False) -> None:
        """Save the statistics to a JSON file."""
        if gzip_file:
            logger.trace(f"Compressing stats file {self.stats_file} with gzip.")
            self.add_stats(
                "output_stats_file",
                str(self.stats_file.with_suffix(".json.gz")),
                "metadata",
            )
            save_stats = {
                k: self._make_json_serializable(v) for k, v in self.stats.items()
            }
            with gzip.open(
                self.stats_file.with_suffix(".json.gz"), "wt", encoding="utf-8"
            ) as f:
                json.dump(save_stats, f, indent=2)
        else:
            logger.trace(f"Writing stats file {self.stats_file} without compression")
            self.add_stats("output_stats_file", str(self.stats_file), "metadata")
            save_stats = {
                k: self._make_json_serializable(v) for k, v in self.stats.items()
            }
            with self.stats_file.open("w") as f:
                json.dump(save_stats, f, indent=2)

    def _make_json_serializable(self, obj):
        """Convert non-JSON serializable objects to serializable forms."""
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, (np.bool_, bool)):
            return bool(obj)
        elif isinstance(obj, dict):
            return {k: self._make_json_serializable(v) for k, v in obj.items()}
        elif isinstance(obj, (list, tuple)):
            return [self._make_json_serializable(item) for item in obj]
        elif hasattr(obj, "__dict__"):
            # For custom objects, try to convert to dict
            return self._make_json_serializable(obj.__dict__)
        else:
            return obj

    def format_stats_for_db(self):
        """Format the stats for database insertion. The output dictionary will have keys matching the table names. This function also has internal functions defined in its scope"""
        formatted_stats: dict[str, list[tuple]] = {}
        nifti_path = str(self.path)
        series_uid = self.stats.get("series_uid", None)
        identifier = {"nifti": nifti_path, "series_uid": series_uid}
        study_uid = self.stats.get("study_uid", None)

        def _format_metadata() -> list[dict]:
            """Format the metadata for database insertion."""
            metadata = self.stats.get("metadata", {})
            if not metadata:
                logger.debug(f"No metadata found in the stats for {self.path}")
                return []
            # Build row data from metadata, handling renamed columns
            row_data = {**metadata}
            row_data["nifti"] = nifti_path
            row_data["series_uid"] = series_uid
            row_data["study_uid"] = study_uid
            # Handle renamed columns (old name -> new name)
            if "birthday" in metadata and "birth_date" not in row_data:
                row_data["birth_date"] = metadata["birthday"]
            if "length_mm" in metadata and "pixel_length_mm" not in row_data:
                row_data["pixel_length_mm"] = metadata["length_mm"]
            if "width_mm" in metadata and "pixel_width_mm" not in row_data:
                row_data["pixel_width_mm"] = metadata["width_mm"]
            if "age_at_scan" in metadata and "age" not in row_data:
                row_data["age"] = metadata["age_at_scan"]
            if "manufacturer_model" in metadata and "model" not in row_data:
                row_data["model"] = metadata["manufacturer_model"]
            return [row_data]

        def _format_vol_int() -> list[dict]:
            "Format the vol_int data for database insertion"
            vol_int = self.stats.get("vol_int", {})
            if not vol_int:
                logger.debug(f"No vol_int stats found in {self.path}")
                return []
            rows = []
            for structure, stats in vol_int.items():
                rows.append(
                    {
                        **identifier,
                        "structure": structure,
                        "volume_cm3": stats.get("volume_cm3"),
                        "hu_mean": stats.get("hu_mean"),
                        "hu_std_dev": stats.get("hu_std_dev"),
                    }
                )
            return rows

        def _format_contrast() -> list[dict]:
            contrast = self.stats.get("contrast", {})
            if not contrast:
                logger.debug(f"No contrast stats found in {self.path}")
                return []
            return [
                {
                    **identifier,
                    "phase": contrast.get("phase"),
                    "probability": contrast.get("probability"),
                    "pi_time": contrast.get("pi_time"),
                    "pi_time_std": contrast.get("pi_time_std"),
                }
            ]

        def _format_vertebrae() -> list[dict]:
            vertebrae = self.stats.get("vertebrae", {})
            if not vertebrae:
                logger.debug(f"No vertebrae stats found in {self.path}")
                return []
            rows = []
            for vertebra, stats in vertebrae.items():
                rows.append(
                    {
                        **identifier,
                        "vertebra": vertebra,
                        "midline": stats.get("midline"),
                    }
                )
            return rows

        def _format_aorta_metrics() -> list[dict]:
            aorta_metrics = self.stats.get("aorta", {})
            kvp = self.stats.get("metadata", {}).get("kvp", None)
            if kvp is None:
                pass
            elif kvp == 120:
                kvp = 1
            else:
                kvp = 0
            if not aorta_metrics:
                logger.debug(f"No aorta metrics found in {self.path}")
                return []
            rows = []
            for region, stats in aorta_metrics.items():
                rows.append(
                    {
                        **identifier,
                        "region": region,
                        "entire_region": stats.get("entire_region"),
                        "length_mm": stats.get("length_mm"),
                        "tortuosity_index": stats.get("tortuosity_index"),
                        "icm": stats.get("icm"),
                        "n_inflections": stats.get("n_inflections"),
                        "peria_volume_cm3": stats.get("peria_volume_cm3"),
                        "peria_ring_volume_cm3": stats.get("peria_ring_volume_cm3"),
                        "peria_fat_volume_cm3": stats.get("peria_fat_volume_cm3"),
                        "peria_hu_mean": stats.get("peria_hu_mean"),
                        "peria_hu_std": stats.get("peria_hu_std"),
                        "calc_volume_mm3": stats.get("calc_volume_mm3"),
                        "calc_agatston": stats.get("calc_agatston"),
                        "calc_count": stats.get("calc_count"),
                        "is_120_kvp": kvp,
                        "mean_diameter_mm": stats.get("mean", {}).get("diameter"),
                        "mean_roundness": stats.get("mean", {}).get("roundness"),
                        "mean_flatness": stats.get("mean", {}).get("flatness"),
                    }
                )
            return rows

        def _format_aorta_diameters() -> list[dict]:
            aorta_metrics = self.stats.get("aorta", {})
            if not aorta_metrics:
                logger.debug(f"No aorta diameters found in {self.path}")
                return []
            rows = []
            measure_keys = ["max", "min", "mid", "proximal", "distal"]
            for region, stats in aorta_metrics.items():
                if region == "whole":
                    continue  # Skip the whole aorta region
                entire_region = stats.get("entire_region")
                for measure in measure_keys:
                    if measure not in stats:
                        continue
                    measure_stats = stats[measure]
                    rows.append(
                        {
                            **identifier,
                            "region": region,
                            "measure": measure,
                            "mean_diameter_mm": measure_stats.get("mean_diameter"),
                            "major_diameter_mm": measure_stats.get("major_diameter"),
                            "minor_diameter_mm": measure_stats.get("minor_diameter"),
                            "area_mm2": measure_stats.get("area"),
                            "roundness": measure_stats.get("roundness"),
                            "flatness": measure_stats.get("flatness"),
                            "rel_distance": measure_stats.get("rel_distance"),
                            "entire_region": entire_region,
                        }
                    )
            return rows

        def _format_tissues_volumetric() -> list[dict]:
            volumetric = self.stats.get("tissues", {}).get("volumetric", {})
            if not volumetric:
                logger.debug(f"No tissue volumetric stats found in {self.path}")
                return []
            rows = []
            for region, stats in volumetric.items():
                for structure, structure_stats in stats.items():
                    rows.append(
                        {
                            **identifier,
                            "region": region,
                            "structure": structure,
                            "volume_cm3": structure_stats.get("volume_cm3"),
                            "hu_mean": structure_stats.get("hu_mean"),
                            "hu_std_dev": structure_stats.get("hu_std_dev"),
                        }
                    )
            return rows

        def _format_tissues_vertebral() -> list[dict]:
            vertebral = self.stats.get("tissues", {}).get("vertebral", {})
            if not vertebral:
                logger.debug(f"No tissue vertebral stats found in {self.path}")
                return []
            rows = []
            for vertebra, stats in vertebral.items():
                for structure, structure_stats in stats.items():
                    for measure, value in structure_stats.items():
                        rows.append(
                            {
                                **identifier,
                                "vertebra": vertebra,
                                "structure": structure,
                                "measurement": measure,
                                "value": value,
                            }
                        )
            return rows

        def _format_iliac() -> list[dict]:
            rows = []
            iliac = self.stats.get("iliac", {})
            if not iliac:
                logger.debug(f"No iliac stats found in {self.path}")
                return rows
            for side, data in iliac.items():
                length_mm = data.pop("length_mm", 0.0)
                for location, metrics in data.items():
                    for metric, value in metrics.items():
                        rows.append(
                            {
                                **identifier,
                                "side": side,
                                "length_mm": length_mm,
                                "location": location,
                                "metric": metric,
                                "value": value,
                            }
                        )
            return rows

        for table in STATS_TABLES:
            match table:
                case "metadata":
                    table_data = _format_metadata()
                case "vol_int":
                    table_data = _format_vol_int()
                case "contrast":
                    table_data = _format_contrast()
                case "vertebrae":
                    table_data = _format_vertebrae()
                case "aorta_metrics":
                    table_data = _format_aorta_metrics()
                case "aorta_diameters":
                    table_data = _format_aorta_diameters()
                case "tissues_volumetric":
                    table_data = _format_tissues_volumetric()
                case "tissues_vertebral":
                    table_data = _format_tissues_vertebral()
                case "iliac":
                    table_data = _format_iliac()
            formatted_stats[table] = table_data
        self.db_stats = formatted_stats
        return self


def resample_with_sitk(
    image,
    new_spacing=None,
    new_size=None,
    is_label=False,
    interpolator_type="gaussian",
    reference_image=None,
    reorient=None,
) -> sitk.Image:
    """Resample a SimpleITK image to a new spacing, size, or reference image.
    :param image: The SimpleITK image to resample.
    :param new_spacing: The new spacing to resample to. If None, the original
        spacing will be used to calculate the new size.
    :param new_size: The new size to resample to. If None, the original
        size will be used to calculate the new spacing.
    :param is_label: Whether the image is a label image. If True, the
        resampling will use label-specific interpolators.
    :param interpolator_type: The type of interpolator to use for resampling.
        Options are 'gaussian', 'linear', 'nearest' for label images, and
        'lanczos', 'bspline', 'gaussian', 'linear' for regular images.
    :param reference_image: An optional reference image to use for resampling.
        If provided, the new spacing and size will be taken from this image.
    :param reorient: The CT orientation to reorient the image to. If None, no reorientation is applied.
    :return: The resampled SimpleITK image."""
    # Get original properties
    original_spacing = image.GetSpacing()
    original_size = image.GetSize()

    resampler = sitk.ResampleImageFilter()
    if reference_image is not None:
        resampler.SetReferenceImage(reference_image)
    else:
        if new_spacing is not None and new_size is not None:
            raise ValueError("Only one of new_spacing and new_size may be specified")
        # Calculate new size based on spacing change
        if new_spacing is not None:
            new_size = [
                int(original_size[i] * original_spacing[i] / new_spacing[i])
                for i in range(3)
            ]
        elif new_size is not None:
            new_spacing = [
                original_spacing[i] * original_size[i] / new_size[i] for i in range(3)
            ]
        else:
            raise ValueError("Must specify either a new_size or new_spacing")
        #    Set up resampling parameters
        resampler.SetOutputSpacing(new_spacing)
        resampler.SetSize(new_size)
        resampler.SetOutputDirection(image.GetDirection())
        resampler.SetOutputOrigin(image.GetOrigin())
        resampler.SetTransform(sitk.Transform())
        resampler.SetDefaultPixelValue(0)

    if is_label:
        if interpolator_type == "gaussian":
            interpolator = sitk.sitkLabelGaussian
        elif interpolator_type == "linear":
            interpolator = sitk.sitkLabelLinear
        elif interpolator_type == "nearest":
            interpolator = sitk.sitkNearestNeighbor
        else:
            raise ValueError(
                f"label interpolator must be in [gaussian, linear, nearest], got {interpolator_type}"
            )
        resampler.SetInterpolator(interpolator)
    else:
        if interpolator_type == "lanczos":
            interpolator = sitk.sitkLanczosWindowedSinc
        elif interpolator_type == "bspline":
            interpolator = sitk.sitkBSpline
        elif interpolator_type == "gaussian":
            interpolator = sitk.sitkGaussian
        elif interpolator_type == "linear":
            interpolator = sitk.sitkLinear
        else:
            raise ValueError(
                f"image interpolator must be in [lanczos, gaussian, linear], got {interpolator_type}"
            )
        resampler.SetInterpolator(interpolator)
    resampled_image = resampler.Execute(image)
    if reorient is not None:
        current_orientation = (
            sitk.DICOMOrientImageFilter_GetOrientationFromDirectionCosines(
                resampled_image.GetDirection()
            )
        )
        if current_orientation != reorient:
            logger.trace(f"Reorienting image from {current_orientation} to {reorient}.")
            reorienter = sitk.DICOMOrientImageFilter()
            reorienter.SetDesiredCoordinateOrientation(reorient)
            resampled_image = reorienter.Execute(resampled_image)
    return resampled_image


def load_nifti_paths(input_arg: Path, data_dir: Path) -> list[Path]:
    """Store either the contents of the input text file or the input nifti as a list of nifti paths
    Parameters
        input_arg: Path - the input argument path
        data_dir: Path - the data directory to prepend to the nifti file name {data_dir} / {nifti_path}
    Returns:
        A list of paths containing the nifti files in the input
    """
    single_file_suffixes = set([".nii", ".gz"])
    # Input arg must exist
    if not input_arg.exists():
        raise FileNotFoundError(
            f"Input argument {input_arg} does not exist. Please check path."
        )
    # If it's a single file - just prepend the data dir and return
    if set(input_arg.suffixes).intersection(single_file_suffixes):
        logger.info(f"Found input nifti {data_dir / input_arg}")
        return [data_dir / input_arg]
    # Otherwise we load the file and validate
    with input_arg.open() as f:
        niftis = [data_dir / Path(line.strip()) for line in f.readlines()]
    if len(niftis) == 0:
        raise ValueError(
            f"Nifti list file {input_arg} is empty. Please check contents."
        )
    logger.info(f"Found {len(niftis)} nifti files to analyze in {input_arg}")
    return niftis


class NotNiftiFileError(ValueError):
    pass


class SimpleITKReadError(ValueError):
    pass



from pathlib import Path


def add_dicom_conversion_subparser(subparsers):
    # Add subcommands
    convert_parser = subparsers.add_parser(
        "convert", help="Convert dicom data to the NIfTI format"
    )
    convert_parser.add_argument(
        "dicoms",
        help="Path to a dicom directory or a file containing a list of dicom directories",
        type=Path,
    )
    convert_parser.add_argument(
        "output_dir",
        type=Path,
        help="Path to the output directory",
    )
    convert_parser.add_argument(
        "--data-dir",
        "-d",
        type=Path,
        default=Path(""),
        help="Data directory for the input dicoms (i.e. {data_dir}/dicom_folder) - useful for remapping in docker. Default is no data directory.",
    )
    convert_parser.add_argument(
        "--n-processes",
        "-n",
        type=int,
        default=1,
        help="Number of processes to use for conversion (default: 1)",
    )
    convert_parser.add_argument(
        "--threads-per-process",
        "-t",
        type=int,
        default=4,
        help="Maximum number of transformation threads per process (default: 4)",
    )
    convert_parser.add_argument(
        "--axial-only",
        "-ax",
        action="store_true",
        help="Only convert axial series",
    )
    convert_parser.add_argument(
        "--no-mip", "-nm", action="store_true", help="Do not convert MIP series"
    )
    convert_parser.add_argument(
        "--no-resampling",
        "-nr",
        action="store_true",
        help="Do not resample the dicoms to LAS orientation",
    )
    convert_parser.add_argument(
        "--resample-interpolation-order",
        "-i",
        type=int,
        default=3,
        help="Interpolation order for resampling (default: 3)",
    )
    convert_parser.add_argument(
        "--resample-padding",
        "-p",
        type=int,
        default=-1024,
        help="Padding Hounsfield Unit for resampling (default: -1024)",
    )
    convert_parser.add_argument(
        "--validate-orthogonal",
        "-v",
        action="store_true",
        help="Validate that the dicoms are orthogonal. Will not convert gantry tilted series.",
    )
    convert_parser.add_argument(
        "--skip-slice-increment-validation",
        "-ssi",
        action="store_true",
        help="Skip validation of consistent slice increment in DICOM series. Use with caution",
    )
    convert_parser.add_argument(
        "--skip-instance-number-validation",
        "-sin",
        action="store_true",
        help="Skip validation of dicom instance numbers. Use with caution.",
    )
    convert_parser.add_argument(
        "--skip-slice-count-validation",
        "-ssc",
        action="store_true",
        help="Skip validation of slice counts in the dicoms. Use with caution.",
    )
    convert_parser.add_argument(
        "--min-slice-count",
        "-s",
        type=int,
        default=30,
        help="Minimum number of slices to consider a series valid (default: 30)",
    )
    convert_parser.add_argument(
        "--db-batch-size",
        "-b",
        type=int,
        default=100,
        help="Batch size for database insertion (default: 100)",
    )
    convert_parser.add_argument(
        "--ignore",
        action="store_true",
        help="Ignore primary key duplicates in database instead of replacing them. Default is to replace.",
    )



import duckdb
from pathlib import Path

import re

from loguru import logger
from mircat_v2.configs import read_dbase_config, write_config

# Path to the DuckDB schema file
dbase_schema_file = Path(__file__).parent / "configs" / "dbase_duckdb.sql"


def add_dbase_subparser(subparsers):
    """Add database management subcommands to the CLI."""
    dbase_parser = subparsers.add_parser(
        "dbase",
        help="Database management commands",
        description="mircat-v2 database management commands.",
    )
    dbase_subparsers = dbase_parser.add_subparsers(
        dest="dbase_command", description="Database operations:"
    )

    # Create a new database
    create_parser = dbase_subparsers.add_parser("create", help="Create a new database")
    create_parser.add_argument(
        "dbase_path",
        type=Path,
        help="Path to the database file",
    )
    create_parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite the database if it already exists",
    )

    # Set an existing database
    set_parser = dbase_subparsers.add_parser(
        "set", help="Set an existing DuckDB database file to be used in mircat-v2"
    )
    set_parser.add_argument(
        "dbase_path", type=Path, help="Path to the existing database file."
    )
    set_parser.add_argument(
        "--create-if-missing",
        "-c",
        action="store_true",
        help="Create the database if it does not exist.",
    )

    # Update an existing database schema
    update_parser = dbase_subparsers.add_parser(
        "update", help="Update an existing database schema to match the current mircat-v2 schema"
    )
    update_parser.add_argument(
        "dbase_path",
        type=Path,
        nargs="?",
        default=None,
        help="Path to the database file. If not provided, uses the configured database.",
    )


def run_dbase_command(args) -> None:
    """Run the appropriate function depending on the given dbase command."""
    match args.dbase_command:
        case "create":
            create_database(args.dbase_path, args.overwrite)
        case "set":
            set_database(args.dbase_path, args.create_if_missing)
        case "update":
            update_database(args.dbase_path)
        case _:
            logger.error(f"Unknown dbase command: {args.dbase_command}")


def create_database(dbase_path: Path, overwrite: bool = False) -> None:
    """Create a new DuckDB database at the specified path.

    Parameters
    ----------
    dbase_path : Path
        Path to the database file.
    overwrite : bool
        If True, overwrite the database if it already exists.
    """
    if dbase_path.exists() and not overwrite:
        raise FileExistsError(
            f"Database file {dbase_path} already exists. Use `mircat-v2 dbase set` to use it or `--overwrite` to completely overwrite."
        )
    elif dbase_path.exists() and overwrite:
        response = input(
            "Are you sure you want to overwrite the existing database? This will not copy data to the new one. (y/n): "
        )
        if response.lower() != "y":
            logger.info(f"Left existing database at {dbase_path} intact.")
            return
        logger.info(f"Overwriting existing database at {dbase_path}")
        dbase_path.unlink()

    conn = duckdb.connect(str(dbase_path))
    _create_tables_from_schema(conn)
    conn.close()
    logger.success(f"Database created at {dbase_path}")
    _save_dbase_path(dbase_path)


def set_database(dbase_path: Path, create_if_missing: bool) -> None:
    """Set the database at the given path to be used for database operations."""
    if not dbase_path.exists() and not create_if_missing:
        raise FileNotFoundError(
            f"The database at {dbase_path} was not found. Use `mircat-v2 dbase create {dbase_path}` to create one."
        )
    if create_if_missing and not dbase_path.exists():
        logger.info(f"Database not found, creating new database at {dbase_path}")
        conn = duckdb.connect(str(dbase_path))
        _create_tables_from_schema(conn)
        conn.close()
        logger.success(f"Database created at {dbase_path}")
    _save_dbase_path(dbase_path)
    logger.success(f"Set {dbase_path} as the database for mircat-v2.")


def update_database(dbase_path: Path | None = None) -> None:
    """Update an existing database schema to match the current mircat-v2 schema.

    This function will:
    1. Add new ENUM values to existing ENUMs
    2. Create new tables that don't exist
    3. Add new columns to existing tables

    Parameters
    ----------
    dbase_path : Path | None
        Path to the database file. If None, uses the configured database.
    """
    if dbase_path is None:
        config = read_dbase_config()
        dbase_path = Path(config["dbase_path"])

    if not dbase_path.exists():
        raise FileNotFoundError(
            f"Database file {dbase_path} not found. Use `mircat-v2 dbase create` to create one."
        )

    logger.info(f"Updating database schema at {dbase_path}")

    conn = duckdb.connect(str(dbase_path))
    try:
        # Parse the schema file
        with dbase_schema_file.open("r") as f:
            schema_script = f.read()

        # Update ENUMs
        _update_enums(conn, schema_script)

        # Update tables (create missing tables and add missing columns)
        _update_tables(conn, schema_script)

        logger.success(f"Database schema updated at {dbase_path}")
    finally:
        conn.close()


def _parse_enum_definitions(schema_script: str) -> dict[str, list[str]]:
    """Parse ENUM definitions from the schema script.

    Returns a dict mapping enum name to list of values.
    """
    enum_pattern = r"CREATE TYPE IF NOT EXISTS (\w+) AS ENUM \(([^)]+)\)"
    enums = {}
    for match in re.finditer(enum_pattern, schema_script):
        enum_name = match.group(1)
        values_str = match.group(2)
        values = [v.strip().strip("'") for v in values_str.split(",")]
        enums[enum_name] = values
    return enums


def _get_existing_enum_values(conn: duckdb.DuckDBPyConnection, enum_name: str) -> list[str]:
    """Get existing values for an ENUM type from the database."""
    try:
        result = conn.execute(
            f"SELECT unnest(enum_range(NULL::{enum_name}))"
        ).fetchall()
        return [row[0] for row in result]
    except duckdb.CatalogException:
        return []


def _update_enums(conn: duckdb.DuckDBPyConnection, schema_script: str) -> None:
    """Update ENUM types to add any new values.

    DuckDB doesn't support ALTER TYPE ... ADD VALUE, so we need to drop and
    recreate ENUMs. This requires finding all columns using the ENUM,
    dropping them, recreating the ENUM, then recreating the columns.
    """
    schema_enums = _parse_enum_definitions(schema_script)

    for enum_name, schema_values in schema_enums.items():
        existing_values = _get_existing_enum_values(conn, enum_name)

        if not existing_values:
            # ENUM doesn't exist, create it
            values_str = ", ".join(f"'{v}'" for v in schema_values)
            conn.execute(f"CREATE TYPE {enum_name} AS ENUM ({values_str})")
            logger.info(f"Created ENUM type: {enum_name}")
            continue

        # Find new values that need to be added
        new_values = [v for v in schema_values if v not in existing_values]

        if new_values:
            # DuckDB doesn't support ALTER TYPE ADD VALUE, so we need to
            # drop and recreate the ENUM along with dependent columns
            _recreate_enum_with_new_values(conn, enum_name, schema_values)
            logger.info(f"Added values to ENUM {enum_name}: {', '.join(new_values)}")


def _get_columns_using_enum(conn: duckdb.DuckDBPyConnection, enum_name: str) -> list[tuple[str, str]]:
    """Find all columns that use a specific ENUM type.

    Returns a list of (table_name, column_name) tuples.
    """
    # Query information_schema for columns with this enum type
    result = conn.execute(f"""
        SELECT table_name, column_name
        FROM information_schema.columns
        WHERE data_type = '{enum_name}'
        AND table_schema = 'main'
    """).fetchall()
    return [(row[0], row[1]) for row in result]


def _recreate_enum_with_new_values(
    conn: duckdb.DuckDBPyConnection, enum_name: str, new_values: list[str]
) -> None:
    """Recreate an ENUM type with new values, preserving data in dependent columns.

    This is necessary because DuckDB doesn't support ALTER TYPE ... ADD VALUE.
    """
    # Find all columns using this enum
    dependent_columns = _get_columns_using_enum(conn, enum_name)

    # For each dependent column, convert to VARCHAR temporarily
    for table_name, col_name in dependent_columns:
        # Add a temporary VARCHAR column
        conn.execute(f"ALTER TABLE {table_name} ADD COLUMN {col_name}_temp VARCHAR")
        # Copy data to temp column (cast enum to varchar)
        conn.execute(f"UPDATE {table_name} SET {col_name}_temp = {col_name}::VARCHAR")
        # Drop the original enum column
        conn.execute(f"ALTER TABLE {table_name} DROP COLUMN {col_name}")

    # Drop the old enum type
    conn.execute(f"DROP TYPE {enum_name}")

    # Create the new enum with all values
    values_str = ", ".join(f"'{v}'" for v in new_values)
    conn.execute(f"CREATE TYPE {enum_name} AS ENUM ({values_str})")

    # Recreate the columns with the new enum type
    for table_name, col_name in dependent_columns:
        # Add the column back with the new enum type
        conn.execute(f"ALTER TABLE {table_name} ADD COLUMN {col_name} {enum_name}")
        # Copy data back from temp column
        conn.execute(f"UPDATE {table_name} SET {col_name} = {col_name}_temp::{enum_name}")
        # Drop the temp column
        conn.execute(f"ALTER TABLE {table_name} DROP COLUMN {col_name}_temp")


def _parse_table_definitions(schema_script: str) -> dict[str, str]:
    """Parse CREATE TABLE statements from the schema script.

    Returns a dict mapping table name to full CREATE TABLE statement.
    """
    tables = {}
    # Split by CREATE TABLE to handle each table separately
    parts = re.split(r"(?=CREATE TABLE IF NOT EXISTS)", schema_script)
    for part in parts:
        match = re.match(r"CREATE TABLE IF NOT EXISTS (\w+)\s*\((.+)\)\s*;?\s*$", part, re.DOTALL)
        if match:
            table_name = match.group(1)
            table_def = match.group(2).strip()
            tables[table_name] = table_def
    return tables


def _parse_columns_from_definition(table_def: str) -> dict[str, str]:
    """Parse column definitions from a table definition string.

    Returns a dict mapping column name to column type/definition.
    """
    columns = {}
    # Split by comma, but be careful with CHECK constraints that contain commas
    parts = []
    depth = 0
    current = ""
    for char in table_def:
        if char == "(":
            depth += 1
        elif char == ")":
            depth -= 1
        elif char == "," and depth == 0:
            parts.append(current.strip())
            current = ""
            continue
        current += char
    if current.strip():
        parts.append(current.strip())

    for part in parts:
        part = part.strip()
        # Skip PRIMARY KEY constraints
        if part.upper().startswith("PRIMARY KEY"):
            continue
        # Skip FOREIGN KEY constraints
        if part.upper().startswith("FOREIGN KEY"):
            continue
        # Parse column definition
        tokens = part.split(None, 1)
        if len(tokens) >= 2:
            col_name = tokens[0]
            col_type = tokens[1]
            columns[col_name] = col_type

    return columns


def _get_existing_tables(conn: duckdb.DuckDBPyConnection) -> list[str]:
    """Get list of existing tables in the database."""
    result = conn.execute(
        "SELECT table_name FROM information_schema.tables WHERE table_schema = 'main'"
    ).fetchall()
    return [row[0] for row in result]


def _get_existing_columns(conn: duckdb.DuckDBPyConnection, table: str) -> dict[str, str]:
    """Get existing columns and their types for a table."""
    result = conn.execute(
        f"SELECT column_name, data_type FROM information_schema.columns WHERE table_name = '{table}'"
    ).fetchall()
    return {row[0]: row[1] for row in result}


def _update_tables(conn: duckdb.DuckDBPyConnection, schema_script: str) -> None:
    """Update tables to add missing tables and columns."""
    schema_tables = _parse_table_definitions(schema_script)
    existing_tables = _get_existing_tables(conn)

    for table_name, table_def in schema_tables.items():
        if table_name not in existing_tables:
            # Create the table
            conn.execute(f"CREATE TABLE {table_name} ({table_def})")
            logger.info(f"Created table: {table_name}")
            continue

        # Table exists, check for missing columns
        schema_columns = _parse_columns_from_definition(table_def)
        existing_columns = _get_existing_columns(conn, table_name)

        for col_name, col_type in schema_columns.items():
            if col_name not in existing_columns:
                # Remove CHECK constraints for ALTER TABLE (DuckDB may not support them in ALTER)
                col_type_clean = re.sub(r"\s*CHECK\s*\([^)]+\)", "", col_type).strip()
                conn.execute(f"ALTER TABLE {table_name} ADD COLUMN {col_name} {col_type_clean}")
                logger.info(f"Added column '{col_name}' to table {table_name}")


def _save_dbase_path(dbase_path: Path) -> None:
    """Save the database path to the config file."""
    dbase_config = str(dbase_path.resolve())
    write_config(dbase_config, "dbase", "dbase_path")


def _create_tables_from_schema(conn: duckdb.DuckDBPyConnection) -> None:
    """Execute the DuckDB schema file to create all tables."""
    with dbase_schema_file.open("r") as f:
        schema_script = f.read()
    conn.execute(schema_script)


def _get_table_columns(conn: duckdb.DuckDBPyConnection, table: str) -> list[str]:
    """Get the column names for a table from the database schema."""
    result = conn.execute(
        f"SELECT column_name FROM information_schema.columns WHERE table_name = '{table}'"
    ).fetchall()
    return [row[0] for row in result]


def insert_data_batch(
    dbase_path: str | Path, table: str, data_records: list[dict], ignore: bool = False
) -> None:
    """Insert a batch of data into the database.

    Parameters
    ----------
    dbase_path : str | Path
        Path to the DuckDB database file.
    table : str
        Name of the table to insert into.
    data_records : list[dict]
        List of dictionaries containing the data to insert.
    ignore : bool
        If True, ignore primary key duplicates instead of replacing them.
        Default is False (replace duplicates).
    """
    if not data_records:
        return

    conn = duckdb.connect(str(dbase_path))
    try:
        # Check if table exists, create schema if not
        tables = [
            row[0]
            for row in conn.execute(
                "SELECT table_name FROM information_schema.tables WHERE table_schema = 'main'"
            ).fetchall()
        ]
        if table not in tables:
            logger.info(f"Table {table} does not exist. Creating schema...")
            _create_tables_from_schema(conn)

        # Get column names from the database schema
        columns = _get_table_columns(conn, table)
        if not columns:
            raise ValueError(f"Table {table} not found in database schema.")

        # Filter records to only include valid columns
        insert_data = [
            {k: v for k, v in record.items() if k in columns} for record in data_records
        ]

        if insert_data:
            cols = list(insert_data[0].keys())
            placeholders = ", ".join(["?" for _ in cols])
            col_names = ", ".join(cols)

            # Use INSERT OR IGNORE to skip duplicates, or INSERT OR REPLACE to update them
            if ignore:
                sql = f"INSERT OR IGNORE INTO {table} ({col_names}) VALUES ({placeholders})"
            else:
                sql = f"INSERT OR REPLACE INTO {table} ({col_names}) VALUES ({placeholders})"

            # Prepare values
            values_list = [[record.get(col) for col in cols] for record in insert_data]

            conn.executemany(sql, values_list)
            logger.success(f"DBase: Inserted {len(values_list)} records into {table}")
    except Exception as e:
        logger.error(f"Error inserting batch into {table}: {e}")
        raise
    finally:
        conn.close()



import os
import torch
import traceback
import SimpleITK as sitk

from multiprocessing import Pool
from datetime import date
from pathlib import Path
from loguru import logger
from shutil import copyfile, rmtree
from nnunetv2.inference.predict_from_raw_data import nnUNetPredictor
from threadpoolctl import threadpool_limits

from mircat_v2.configs import read_models_config, logger_setup
from mircat_v2.dbase import insert_data_batch
from mircat_v2.nifti import SegNifti, S3SegNifti, resample_with_sitk


class MircatSegmentor:
    def __init__(
        self,
        niftis: list[Path],
        task_list: list[str],
        model_types: list[str],
        device: str,
        task_configs: dict | None = None,
        n_processes: int = 1,
        threads_per_process: int = 4,
        cache_size: int = 1,
        dbase_config: dict = {},
        img_resampler: str = "bspline",
        lbl_resampler: str = "gaussian",
        verbose: bool = False,
        quiet: bool = False,
        ignore: bool = False,
    ):
        """Initialize segmentor tasks.
        :param task_list: list of all desired segmentation tasks
        :param model_types: list of task model types [3d, 2d, etc.]'
        :param device: the device for pytorch to use
        :param task_configs: dictionary containing task configs. If not passed, will load from library.
        :param n_processes: the number of workers for preprocessing, nnUNet, and postprocessing
        :param threads_per_process: the number of threads each worker can use.
        :param cache_size: Number of nifti files to work on at one time. Default = 1.
        :param dbase_config: dictionary containing database configuration. If empty, no database insertion will be performed.
        :param img_resampler: the type of resampling to use for images, default is "bspline"
        :param lbl_resampler: the type of resampling to use for labels, default is "gaussian"
        """
        self.niftis = niftis
        self.task_list = task_list
        self.model_types = model_types
        self.device = device
        self.n_processes = n_processes
        self.threads_per_process = threads_per_process
        self.cache_size = cache_size
        self.dbase_config = dbase_config
        self.dbase_insert = not dbase_config == {}
        self.img_resampler = img_resampler
        self.lbl_resampler = lbl_resampler
        self.verbose = verbose
        self.quiet = quiet
        self.ignore = ignore
        # Logic for loading task configurations
        if isinstance(task_configs, dict):
            self.task_configs = task_configs
        elif task_configs is None:
            self.task_configs = read_models_config()
        else:
            message = "task_configs argument must be a dictionary or None."
            logger.error(message)
            raise ValueError(message)
        # Get the voxel spacings for each task/model type pair
        self.task_spacings = {
            task: tuple(self.task_configs[task]["models"][model]["spacing"])
            for task, model in zip(task_list, model_types)
        }

        self.task_model_paths = {
            task: self.task_configs[task]["models"][model]["path"]
            for task, model in zip(task_list, model_types)
        }

    def run(self):
        """Summary method to run all specified tasks on all given niftis
        :param niftis: path to either a single nifti file or a text file containing paths (preferably absolute paths) to all niftis for segmentation.
        """
        self._get_nifti_batches()
        for i, nifti_batch in enumerate(self.nifti_batches, start=1):
            self.current_batch = i
            self.run_batch(nifti_batch)
        return

    def run_batch(self, nifti_batch: list[Path]):
        good_niftis, bad_niftis = self._load_nifti_batch(nifti_batch)
        self.niftis_to_segment = good_niftis
        self.failed_niftis = bad_niftis
        if len(good_niftis) == 0:
            logger.warning(
                f"Segmentation [{self.current_batch}/{self.total_batches}] ⚠ No valid niftis found, skipping"
            )
            return
        self._preprocess_batch()
        self._run_tasks()
        self._postprocess_batch()
        if self.dbase_insert:
            self._insert_batch_to_dbase()
        logger.success(
            f"Segmentation [{self.current_batch}/{self.total_batches}] ✓ Batch complete"
        )
        # logic for uploading bad_nifti results to dbase

    def _load_nifti_batch(self, nifti_batch: list[Path]):
        """Load niftis in the batch and filter out bad ones
        :param nifti_batch: the list of nifti paths
        """
        good_niftis = []
        bad_niftis = []
        for nifti_path in nifti_batch:
            try:
                nifti = SegNifti(nifti_path)
                good_niftis.append(nifti)
            except Exception as e:
                bad_niftis.append(
                    {
                        "nifti": str(nifti_path),
                        "seg_date": date.today().isoformat(),
                        "status": "FAILED",
                        "failed_error": type(e).__name__,
                    }
                )
        logger.debug(
            f"Found {len(good_niftis)} niftis able to be loaded by SimpleITK for segmentation, failed to load {len(bad_niftis)} niftis."
        )
        return good_niftis, bad_niftis

    def _preprocess_batch(self) -> None:
        """Preprocess each nifti in a batch by writing each necessary temporary file
        :param good_niftis: the list of pre-validated Niftis instances from _load_nifti_batch
        """
        preprocessed = []
        logger.info(
            f"Segmentation - Preprocessing: [{self.current_batch}/{self.total_batches}] Starting"
        )
        with Pool(self.n_processes) as pool:
            for i, nifti in enumerate(
                pool.imap_unordered(self._preprocess_nifti, self.niftis_to_segment),
                start=1,
            ):
                preprocessed.append(nifti)
                logger.info(
                    f"Segmentation - Preprocessing: [{self.current_batch}/{self.total_batches}] ({i}/{len(self.niftis_to_segment)})"
                )
        logger.success(
            f"Segmentation - Preprocessing: [{self.current_batch}/{self.total_batches}] ✓ Complete"
        )
        self.niftis_to_segment = preprocessed

    def _preprocess_nifti(self, nifti: SegNifti) -> SegNifti:
        """Preprocess a single nifti file for segmentation by creating the necessary spacing files and storing the information
        :param nifti: An instance of the Nifti class
        """
        logger_setup(self.verbose, self.quiet)
        limits = threadpool_limits(self.threads_per_process)
        sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(self.threads_per_process)
        with limits:
            if not nifti.seg_folder.exists():
                logger.debug("Making segmentation folder {}", nifti.seg_folder)
                nifti.seg_folder.mkdir()
            else:
                logger.debug(
                    "Segmentation folder aready found for {}", nifti.seg_folder
                )
            preprocessed_files = {}
            for task, spacing in self.task_spacings.items():
                spacing_str: str = "_".join([str(x) for x in spacing])
                preprocessed_file: Path = (
                    nifti.seg_folder / f"{nifti.name}_{spacing_str}.nii.gz"
                )
                output_file: Path = (
                    nifti.seg_folder / f"{nifti.name}_{spacing_str}_{task}"
                )
                if not preprocessed_file.exists():
                    logger.debug(f"Making {preprocessed_file} for task {task}..")
                    nifti.resample_and_save_for_segmentation(
                        spacing, preprocessed_file, self.img_resampler
                    )
                else:
                    logger.debug(
                        f"Preprocessed file {preprocessed_file} already exists for task {task}."
                    )
                preprocessed_files[task] = {
                    "input": preprocessed_file,
                    "output": output_file,
                }
            nifti.task_files = preprocessed_files
            return nifti

    def _get_nifti_batches(self) -> None:
        """Ensure that the input niftis are in the correct format. If the input is a text file, we read the contents to get the paths of all niftis."""
        self.nifti_batches = [
            self.niftis[i : i + self.cache_size]
            for i in range(0, len(self.niftis), self.cache_size)
        ]
        self.total_batches = len(self.nifti_batches)
        logger.info(
            f"Running tasks {self.task_list} on {len(self.nifti_batches)} batches of {self.cache_size} niftis per batch."
        )

    def _run_tasks(self):
        # if self.device != "cpu":
        #     # nnUNet compile only seems to want to work with GPUs
        #     os.environ["nnUNet_compile"] = "True"
        os.environ["nnUNet_compile"] = "False"
        device = torch.device(self.device)
        logger.info("Running on torch.device {}", device)
        for task in self.task_list:
            logger.info("Starting task {} with nnUNet", task)
            input_files = [
                [str(nifti.task_files.get(task).get("input"))]
                for nifti in self.niftis_to_segment
            ]
            output_files = [
                str(nifti.task_files.get(task).get("output"))
                for nifti in self.niftis_to_segment
            ]
            predictor = nnUNetPredictor(
                tile_step_size=0.5,
                use_gaussian=True,
                use_mirroring=False,
                perform_everything_on_device=True if self.device != "cpu" else False,
                device=device,
                verbose=False,
                verbose_preprocessing=False,
                allow_tqdm=True,
            )
            model_path = self.task_model_paths.get(task)
            if not isinstance(model_path, str) or not model_path:
                raise ValueError(
                    f"Model path for task '{task}' is missing or not a string."
                )
            predictor.initialize_from_trained_model_folder(
                model_path,
                use_folds=(0,),
                checkpoint_name="checkpoint_final.pth",
            )
            predictor.predict_from_files(
                input_files,
                output_files,
                save_probabilities=False,
                overwrite=True,
                num_processes_preprocessing=self.n_processes,
                num_processes_segmentation_export=self.n_processes,
            )
            del predictor

    def _postprocess_batch(self):
        n_niftis = len(self.niftis_to_segment)
        completed = []
        logger.info(
            f"Segmentation - Postprocessing: [{self.current_batch}/{self.total_batches}] Starting"
        )
        with Pool(self.n_processes) as pool:
            for i, dbase_insert in enumerate(
                pool.imap_unordered(self._postprocess_nifti, self.niftis_to_segment),
                start=1,
            ):
                completed.extend(dbase_insert)
                logger.info(
                    f"Segmentation - Postprocessing: [{self.current_batch}/{self.total_batches}] ({i}/{n_niftis})"
                )
        self.completed_niftis = completed

    def _postprocess_nifti(self, nifti: SegNifti):
        logger_setup(self.verbose, self.quiet)
        sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(self.threads_per_process)
        dbase_insert = []
        with threadpool_limits(self.threads_per_process):
            for task, task_files in nifti.task_files.items():
                input_nifti = task_files["input"]
                tmp_output_nifti = Path(f"{task_files['output']}.nii.gz")
                resampled_name = nifti.seg_folder / f"{nifti.name}_{task}.nii.gz"
                # Delete the preprocessed input
                if input_nifti.exists():
                    input_nifti.unlink()
                # Resample the output and create the new output file, delete the old after
                if tmp_output_nifti.exists():
                    lbl = sitk.ReadImage(str(tmp_output_nifti))
                    new_lbl = resample_with_sitk(
                        lbl,
                        new_size=nifti.shape,
                        is_label=True,
                        interpolator_type=self.lbl_resampler,
                    )
                    sitk.WriteImage(new_lbl, resampled_name)
                    logger.debug("successfully wrote {}", resampled_name)
                    tmp_output_nifti.unlink()
                else:
                    logger.warning(f"Output file {tmp_output_nifti} not found!")
                dbase_task = {
                    "nifti": str(nifti.path),
                    "series_uid": nifti.metadata.get("series_uid"),
                    "task": int(task),
                    "seg_file": str(resampled_name),
                    "seg_date": date.today().isoformat(),
                    "status": "SUCCESS",
                    "failed_error": None,
                }
                dbase_insert.append(dbase_task)
        return dbase_insert

    def _insert_batch_to_dbase(self):
        logger.info(
            f"Segmentation [{self.current_batch}/{self.total_batches}] Inserting results into database"
        )
        insert_data_batch(self.dbase_config["dbase_path"], "segmentations", self.completed_niftis, self.ignore)
        logger.debug(
            f"Segmentation [{self.current_batch}/{self.total_batches}] Completed niftis inserted"
        )
        insert_data_batch(self.dbase_config["dbase_path"], "segmentations", self.failed_niftis, self.ignore)
        logger.debug(
            f"Segmentation [{self.current_batch}/{self.total_batches}] Failed niftis inserted"
        )
        logger.success(
            f"Segmentation [{self.current_batch}/{self.total_batches}] ✓ Results inserted into database"
        )


class S3Segmentor(MircatSegmentor):
    def run(self, temp_dir: Path):
        """Summary method to run all specified tasks on all given niftis
        Parameters:
            temp_dir: Path - the temporary directory to use for downloading and storing nifti files
        """
        self._get_nifti_batches()
        for i, nifti_batch in enumerate(self.nifti_batches, start=1):
            self.current_batch = i
            self.run_batch(nifti_batch, temp_dir)
        logger.success("All batches complete!")
        exit(0)

    def run_batch(self, nifti_batch: list[Path], temp_dir: Path):
        good_niftis, bad_niftis = self._load_nifti_batch(nifti_batch, temp_dir)
        self.niftis_to_segment = good_niftis
        self.failed_niftis = bad_niftis
        if len(good_niftis) == 0:
            logger.warning(
                f"Segmentation [{self.current_batch}/{self.total_batches}] ⚠ No valid niftis found, skipping"
            )
            return
        self._preprocess_batch()
        self._run_tasks()
        self._postprocess_batch()
        self._clear_temp_dir(temp_dir)
        if self.dbase_insert:
            self._insert_batch_to_dbase()
        logger.success(
            f"Segmentation [{self.current_batch}/{self.total_batches}] ✓ Batch complete"
        )
        # logic for uploading bad_nifti results to dbase

    def _load_nifti_batch(self, nifti_batch: list[Path], temp_dir: Path):
        """Load niftis in the batch and filter out bad ones
        Parameters:
            nifti_batch: the list of nifti paths
            temp_dir: the temporary directory to use for downloading and storing nifti files
        """
        good_niftis = []
        bad_niftis = []
        for nifti_path in nifti_batch:
            logger.debug("working on {} from s3", nifti_path)
            try:
                nifti = S3SegNifti(nifti_path, temp_dir)
                good_niftis.append(nifti)
            except Exception as e:
                bad_niftis.append(
                    {
                        "nifti": str(nifti_path),
                        "seg_date": date.today().isoformat(),
                        "status": "FAILED",
                        "failed_error": type(e).__name__,
                    }
                )
        logger.debug(
            f"Found {len(good_niftis)} niftis able to be copied from s3 and loaded by SimpleITK for segmentation, failed to load {len(bad_niftis)} niftis."
        )
        return good_niftis, bad_niftis

    def _clear_temp_dir(self, temp_dir: Path):
        """Clear the temporary directory used for storing nifti files after segmentation is complete
        Parameters:
            the temporary directory to use for downloading and storing nifti files
        """
        for item in temp_dir.iterdir():
            try:
                if item.is_dir():
                    rmtree(item)
                else:
                    item.unlink()
            except Exception as e:
                logger.error(f"Error deleting {item}: {e}, {traceback.format_exc()}")
        logger.success(
            f"Segmentation [{self.current_batch}/{self.total_batches}] ✓ Temporary directory cleared"
        )

    def _postprocess_batch(self):
        n_niftis = len(self.niftis_to_segment)
        completed = []
        logger.info(
            f"Segmentation - Postprocessing: [{self.current_batch}/{self.total_batches}] Starting"
        )
        with Pool(self.n_processes) as pool:
            for i, dbase_insert in enumerate(
                pool.imap_unordered(self._postprocess_nifti, self.niftis_to_segment),
                start=1,
            ):
                completed.extend(dbase_insert)
                logger.info(
                    f"Segmentation - Postprocessing: [{self.current_batch}/{self.total_batches}] ({i}/{n_niftis})"
                )
        self.completed_niftis = completed

    def _postprocess_nifti(self, nifti):
        logger_setup(self.verbose, self.quiet)
        dbase_insert = []
        with threadpool_limits(self.threads_per_process):
            for task, task_files in nifti.task_files.items():
                input_nifti = task_files["input"]
                tmp_output_nifti = Path(f"{task_files['output']}.nii.gz")
                resampled_name = (
                    nifti.original_seg_folder / f"{nifti.name}_{task}.nii.gz"
                )
                # Delete the preprocessed input
                if input_nifti.exists():
                    input_nifti.unlink()
                # Copy the segmented file to the original location
                if not nifti.original_seg_folder.exists():
                    nifti.original_seg_folder.mkdir()
                if tmp_output_nifti.exists():
                    logger.debug("copying {} to {}", tmp_output_nifti, resampled_name)
                    copyfile(tmp_output_nifti, resampled_name)

                dbase_task = {
                    "nifti": str(nifti.original_path),
                    "series_uid": nifti.metadata.get("series_uid"),
                    "task": int(task),
                    "seg_file": str(resampled_name),
                    "seg_date": date.today().isoformat(),
                    "status": "SUCCESS",
                    "failed_error": None,
                }
                dbase_insert.append(dbase_task)
        return dbase_insert



import numpy as np
import polars as pl
import SimpleITK as sitk
from multiprocessing import Pool
from pathlib import Path
from loguru import logger
from radiomics import featureextractor, setVerbosity
from threadpoolctl import threadpool_limits
from mircat_v2.nifti import Nifti
from mircat_v2.stats.utilities import calculate_shape_stats
from mircat_v2.configs import read_stats_models_config, read_dbase_config, radiomics_config, logger_setup
from mircat_v2.dbase import insert_data_batch

stats_configs = read_stats_models_config()
allowed_structures = set(["abd_aorta"])
structure_to_model_map = {"abd_aorta": "999"}
structure_to_label_map = {"abd_aorta": 50}
for model_id, model in stats_configs.items():
    for structure, label in model.items():
        if structure != "background":
            allowed_structures.add(structure)
            structure_to_model_map[structure] = model_id
            structure_to_label_map[structure] = label


class RadiomicsExtractor:
    # This does not change across instances
    structure_to_model_map = structure_to_model_map
    structure_to_label_map = structure_to_label_map

    # Initialize with custom labels and config path
    def __init__(
        self,
        structures,
        config_path=None,
        save_to_database=False,
        overwrite=False,
        ignore=False,
        n_processes=1,
        threads_per_process=4,
        verbose=False,
        quiet=False,
    ):
        if config_path is None:
            self.config_path = radiomics_config
        else:
            self.config_path = config_path

        if isinstance(structures, list):
            pass
        elif Path(structures).exists():
            with open(structures) as f:
                structures = f.read().splitlines()
        elif "," in structures:
            structures = structures.split(",")
        elif structures == "all":
            structures = list(allowed_structures)
        else:
            structures = [structures]
        self.structures = self._validate_structures(structures)
        self.save_to_database = save_to_database
        if save_to_database:
            self.dbase_config = read_dbase_config()
        else:
            self.dbase_config = {}
        self.overwrite = overwrite
        self.ignore = ignore
        self.n_processes = n_processes
        self.threads_per_process = threads_per_process
        self.verbose = verbose
        self.quiet = quiet

    @staticmethod
    def _validate_structures(structures) -> list[str]:
        """Validate that given structures are in the allowed structures set.
        Parameters
        ----------
        structures : list[str] - the list of structures to validate against the allowed structures set.

        Returns
        -------
        list[str] - the validated list of structures that are allowed.
        """
        invalid = set(structures).difference(allowed_structures)
        if invalid:
            logger.warning(
                f"Invalid labels detected: {invalid}. Allowed labels are: {allowed_structures}"
            )
            structures = list(set(structures).difference(invalid))
        logger.info(f"Extracting radiomics for {len(structures)} structure{'s' if len(structures) != 1 else ''} from CT.")
        logger.debug(f"Structures: {structures}")
        return structures

    def _get_existing_structures(self, nifti_obj: Nifti) -> set[str]:
        """Return set of structures already extracted for this nifti."""
        output_file = nifti_obj.seg_folder / f"{nifti_obj.name}_radiomics.parquet"
        if not output_file.exists():
            return set()
        structures = (
            pl.scan_parquet(output_file)
            .select("structure")
            .unique()
            .collect()
            .get_column("structure")
            .to_list()
        )
        return set(structures)

    def extract_radiomics(self, niftis: list[Path]):
        """Extract the radiomic features from a list of NIfTI files.
        Parameters
        ----------
        niftis : list[Path]
            List of paths to NIfTI files from which to extract radiomic features.
        """
        self.total_niftis = len(niftis)
        if self.n_processes > 1:
            logger.info(f"Starting radiomics extraction on {self.total_niftis} nifti files with {self.n_processes} processes.")
        else:
            logger.info(f"Starting radiomics extraction on {self.total_niftis} nifti files.")

        if self.n_processes > 1:
            with Pool(processes=self.n_processes) as pool:
                for i, result in enumerate(
                    pool.imap_unordered(self._process_single_nifti, niftis), start=1
                ):
                    if result is None:
                        logger.error(
                            f"Radiomics [{i}/{self.total_niftis}] ({i / self.total_niftis:.2%}) ✗"
                        )
                    else:
                        nifti_name, radiomics_df, structures_extracted = result
                        if self.save_to_database and radiomics_df is not None:
                            self._insert_radiomics_to_database(radiomics_df, structures_extracted)
                        logger.success(
                            f"Radiomics [{i}/{self.total_niftis}] ({i / self.total_niftis:.2%}) ✓ {nifti_name}"
                        )
        else:
            for i, nifti in enumerate(niftis, start=1):
                result = self._process_single_nifti(nifti)
                if result is None:
                    logger.error(
                        f"Radiomics [{i}/{self.total_niftis}] ({i / self.total_niftis:.2%}) ✗"
                    )
                else:
                    nifti_name, radiomics_df, structures_extracted = result
                    if self.save_to_database and radiomics_df is not None:
                        self._insert_radiomics_to_database(radiomics_df, structures_extracted)
                    logger.success(
                        f"Radiomics [{i}/{self.total_niftis}] ({i / self.total_niftis:.2%}) ✓ {nifti_name}"
                    )

    def _process_single_nifti(self, nifti_path: Path) -> tuple | None:
        """Process a single NIfTI file for radiomics extraction. Worker method for multiprocessing.

        Parameters
        ----------
        nifti_path : Path
            Path to the NIfTI file to process.

        Returns
        -------
        tuple | None
            (nifti_name, radiomics_df, structures_extracted) on success, None on failure.
        """
        # Re-initialize logger and limit threads in child processes
        logger_setup(self.verbose, self.quiet)
        sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(self.threads_per_process)
        with threadpool_limits(self.threads_per_process):
            try:
                nifti_obj = Nifti(nifti_path)
                structures_to_extract = self.structures.copy()

                if not self.overwrite:
                    existing = self._get_existing_structures(nifti_obj)
                    structures_to_extract = [
                        s for s in structures_to_extract if s not in existing
                    ]
                    skipped = existing.intersection(set(self.structures))
                    if skipped:
                        logger.debug(
                            f"Skipping already-extracted structures for {nifti_obj.name}: {skipped}"
                        )

                if not structures_to_extract:
                    return (nifti_obj.name, None, [])

                extracted = self._extract_radiomics(nifti_obj, structures_to_extract)
                radiomics_df = self.save_radiomics_to_file(nifti_obj, extracted)
                return (nifti_obj.name, radiomics_df, structures_to_extract)
            except Exception as e:
                logger.error(f"Error processing {nifti_path}: {e}")
                return None

    def _extract_radiomics(self, nifti_obj: Nifti, structures: list[str]) -> None:
        """Radiomics extraction from a nifti file. Saves the extracted features to a file in the segmentation folder.
        Parameters
        ----------
        nifti : Path
            Path to the NIfTI file from which to extract radiomic features.
        structures : list[str]
            List of structures to extract radiomic features for.
        returns:
            None
        """
        segmentations = nifti_obj.segmentations
        # This is the map for older versions of the same model to mapped to the same config
        old_to_new_map = {"total": "999", "tissues": "481", "body": "299", "998": "999"}
        # Get the id map
        id_to_segmentation_map = {}
        for seg in segmentations:
            seg_id = seg.parts[-1].split("_")[-1].split(".")[0]
            seg_id = old_to_new_map.get(seg_id, seg_id)
            id_to_segmentation_map[seg_id] = sitk.ReadImage(seg)
        # Abdominal Aorta is a special
        if "abd_aorta" in structures:
            structures = [
                structure for structure in structures if structure != "abd_aorta"
            ]
            abd_aorta = True
        else:
            abd_aorta = False
        # Read in the image
        setVerbosity(40)
        extractor = featureextractor.RadiomicsFeatureExtractor(str(self.config_path))
        image = nifti_obj.img
        radiomics_output = {}
        for structure in structures:
            seg_id = structure_to_model_map.get(structure)
            if seg_id in id_to_segmentation_map:
                segmentation = id_to_segmentation_map[seg_id]
                label = structure_to_label_map[structure]
                radiomics_output[structure] = {
                    k: v
                    for k, v in extractor.execute(
                        image, segmentation, label=label
                    ).items()
                    if "diagnostics" not in k
                }
        if not abd_aorta:
            return radiomics_output
        # TODO Add abdominal aorta extraction
        seg_id = structure_to_model_map.get("abd_aorta")
        if seg_id in id_to_segmentation_map:
            segmentation = id_to_segmentation_map[seg_id]
            shape_stats = calculate_shape_stats(segmentation)
            if structure_to_label_map.get('abd_aorta') not in shape_stats.GetLabels():
                logger.debug('No abdominal aorta found in this segmentation')
                return radiomics_output
            vertebrae_labels = {
                k.replace("vertebrae_", ""): v
                for k, v in structure_to_label_map.items()
                if "vertebrae" in k.lower()
            }
            vertebrae_stats = {}
            previous_midline = 0
            correct_vertebrae_order = 1
            for vertebra, label in vertebrae_labels.items():
                if label in shape_stats.GetLabels():
                    vertebra_indicies = shape_stats.GetIndexes(label)
                    z_indices = vertebra_indicies[2::3]
                    midline = int(np.median(z_indices))
                    vertebrae_stats[vertebra] = midline
                    if midline <= previous_midline:
                        correct_vertebrae_order = 0
                else:
                    vertebrae_stats[vertebra] = None
            logger.debug(f'Vertebrae indices: {vertebrae_stats}')
            # L5 is lower in the index than T12
            if correct_vertebrae_order == 0:
                logger.debug("Vertebrae order is incorrect. Abdominal aorta extraction may be inaccurate.")
            abd_verts = [v for k, v in vertebrae_stats.items() if v is not None and k in ['L5', 'L4', 'L3', 'L2', 'L1', 'T12']]
            lowest = abd_verts[0] if abd_verts else None
            highest = abd_verts[-1] if abd_verts else None
            l5 = vertebrae_stats.get("L5")
            t12 = vertebrae_stats.get("T12")
            # Track if the abdominal aorta is partial or whole
            whole = 1.0
            if len(abd_verts) < 3:
                logger.warning(f'Not enough vertebrae found to extract abdominal aorta: {nifti_obj.path}')
                return radiomics_output
            elif l5 is None and t12 is None:
                logger.warning(f'Abdominal aorta is not available: {nifti_obj.path}')
                return radiomics_output
            if l5 is None:
                logger.warning(
                    f"Could not find L5 vertebra in {nifti_obj.path}. "
                )
                whole = 0
            elif t12 is None:
                logger.warning(
                    f"Could not find T12 vertebra in {nifti_obj.path}. "
                )
                whole = 0
            image = image[..., lowest:highest]
            segmentation = segmentation[..., lowest:highest]
            label = structure_to_label_map["abd_aorta"]
            radiomics_output["abd_aorta"] = {
                k: v
                for k, v in extractor.execute(image, segmentation, label=label).items()
                if "diagnostics" not in k
            }
            radiomics_output['abd_aorta'].update({'original_shape_is-whole': whole})
        return radiomics_output

    def save_radiomics_to_file(
        self, nifti_obj: Nifti, extracted_radiomics: dict
    ) -> dict:
        """Save extracted radiomics to a file in the segmentations directory of the nifti. Will check for previous radiomics before writing.
        Radiomics file will be saved to {nifti.seg_folder} /  {nifti.name}_radiomics.parquet

        Parameters
        ----------
        nifti_obj:
            the Nifti object containing metadata information
        extracted_radiomics:
            the output of _extract_radiomics

        Returns
        -------
        The dataframe that is saved to a parquet file.
        """
        output_file = Path(f"{nifti_obj.seg_folder}/{nifti_obj.name}_radiomics.parquet")
        if output_file.exists():
            radiomics_df = pl.read_parquet(output_file)
            # If overwriting, remove existing rows for structures being re-extracted
            if self.overwrite:
                structures_to_replace = set(extracted_radiomics.keys())
                radiomics_df = radiomics_df.filter(
                    ~pl.col("structure").is_in(structures_to_replace)
                )
        else:
            radiomics_df = pl.DataFrame(
                schema={
                    "nifti": pl.String,
                    "series_uid": pl.String,
                    "mrn": pl.String,
                    "accession": pl.String,
                    "series_name": pl.String,
                    "series_number": pl.Int64,
                    "scan_date": pl.Date,
                    "structure": pl.String,
                    "transformation": pl.String,
                    "feature_class": pl.String,
                    "feature_name": pl.String,
                    "feature_value": pl.Float64,
                }
            )
        nifti_metadata = nifti_obj.metadata
        nifti_path = str(nifti_obj.path)
        series_uid = nifti_metadata.get("series_uid")
        mrn = nifti_metadata.get("mrn")
        accession = nifti_metadata.get("accession")
        series_name = nifti_metadata.get("series_name")
        series_number = nifti_metadata.get("series_number")
        scan_date = nifti_metadata.get("scan_date", "1900-1-1")
        # Aggregate all radiomics data into a dataframe
        extracted_rows = [
            (
                nifti_path,
                series_uid,
                mrn,
                accession,
                series_name,
                series_number,
                scan_date,
                structure,
                sub_key,
                value,
            )
            for structure, sub_dict in extracted_radiomics.items()
            for sub_key, value in sub_dict.items()
        ]
        extracted_df = (
            pl.DataFrame(
                extracted_rows,
                schema={
                    "nifti": pl.String,
                    "series_uid": pl.String,
                    "mrn": pl.String,
                    "accession": pl.String,
                    "series_name": pl.String,
                    "series_number": pl.Int64,
                    "scan_date": pl.String,
                    "structure": pl.String,
                    "feature_id": pl.String,
                    "feature_value": pl.Float64,
                },
                orient="row",
            )
            .with_columns(
                pl.col("feature_id")
                .str.split_exact("_", 3)
                .struct.rename_fields(
                    ["transformation", "feature_class", "feature_name"]
                ),
                pl.col("scan_date").cast(pl.Date),
            )
            .unnest("feature_id")
            .select(radiomics_df.columns)
        )
        radiomics_df = (
            radiomics_df.vstack(extracted_df)
            .unique()
            .sort("structure", "transformation", "feature_class")
        )
        radiomics_df.write_parquet(output_file)
        return radiomics_df

    def _insert_radiomics_to_database(
        self, radiomics_df: pl.DataFrame, structures: list[str]
    ) -> None:
        """Insert extracted radiomics data into the database.

        Parameters
        ----------
        radiomics_df : pl.DataFrame
            The full radiomics dataframe (may contain other structures too).
        structures : list[str]
            The structures that were just extracted and should be inserted.
        """
        # Filter to only the structures that were just extracted
        # df_to_insert = radiomics_df.filter(pl.col("structure").is_in(structures))
        # if df_to_insert.is_empty():
        #     return

        # Convert scan_date to string for database insertion
        records = radiomics_df.with_columns(
            pl.col("scan_date").cast(pl.String)
        ).to_dicts()
        insert_data_batch(self.dbase_config["dbase_path"], "radiomics", records, self.ignore)


from pathlib import Path
from mircat_v2.radiomics.extractor import RadiomicsExtractor


def add_radiomics_subparser(subparsers):
    # Add subcommands
    radiomics_parser = subparsers.add_parser(
        "radiomics", help="Radiomics feature extraction commands"
    )
    radiomics_parser.add_argument(
        "niftis",
        type=Path,
        help="Path to the input image file or text file containing multiple paths",
    )
    radiomics_parser.add_argument(
        "--data-dir",
        "-d",
        type=Path,
        default=Path(""),
        help="Data directory for the input niftis (i.e. {data_dir}/nifti_path) - useful for remapping in docker. Default is no data directory.",
    )
    radiomics_parser.add_argument(
        "--labels",
        "-l",
        type=str,
        help="Label to extract features for, or a comma-separated list of labels. Default is all labels.",
        default="all",
    )
    radiomics_parser.add_argument(
        "--config",
        "-c",
        type=Path,
        help="Path to a custom YAML configuration file for feature extraction parameters. Runs CT default otherwise.",
        default=None,
    )
    radiomics_parser.add_argument(
        "--database-insert",
        "-db",
        action="store_true",
        help="Flag to store the results of the radiomics extraction in your MirCAT-V2 database",
    )
    radiomics_parser.add_argument(
        "--overwrite",
        "-o",
        action="store_true",
        help="Re-extract and overwrite existing radiomics for specified structures. Default skips already-extracted structures.",
    )
    radiomics_parser.add_argument(
        "--ignore",
        action="store_true",
        help="Ignore primary key duplicates in database instead of replacing them. Default is to replace.",
    )
    radiomics_parser.add_argument(
        "--n-processes",
        "-n",
        type=int,
        default=1,
        help="Number of parallel processes for radiomics extraction. Default is 1 (sequential).",
    )
    radiomics_parser.add_argument(
        "--threads-per-process",
        "-t",
        type=int,
        default=4,
        help="Number of threads each process can use for SimpleITK/NumPy operations. Default is 4.",
    )
    return radiomics_parser


def run_radiomics(args):
    from mircat_v2.nifti import load_nifti_paths

    niftis = load_nifti_paths(args.niftis, args.data_dir)
    extractor = RadiomicsExtractor(
        args.labels,
        args.config,
        args.database_insert,
        args.overwrite,
        args.ignore,
        n_processes=args.n_processes,
        threads_per_process=args.threads_per_process,
        verbose=args.verbose,
        quiet=args.quiet,
    )
    extractor.extract_radiomics(niftis)
